#include "nam-helper.h"
#include "ns3/object.h"
#include "ns3/node-list.h"
#include "ns3/node.h"
#include "ns3/point-to-point-channel.h"
#include "ns3/point-to-point-net-device.h"
#include "ns3/csma-channel.h"
#include "ns3/csma-net-device.h"
#include "ns3/config.h"
#include "ns3/simulator.h"
#include "ns3/log.h"
#include <iostream>
#include <fstream>

NS_LOG_COMPONENT_DEFINE ("NamHelper");

namespace ns3 {

class NamTracer : public Object
{
public:
  NamTracer ();
  ~NamTracer ();

  void StartTracing (std::string filename, double width, double height);
  
private:
  void SetupP2pTraces (uint32_t nodeid, uint32_t deviceid);
  void P2pEnqueueEvent (std::string path, Ptr<const Packet> packet);
  void P2pDequeueEvent (std::string path, Ptr<const Packet> packet);
  void P2pDropEvent (std::string path, Ptr<const Packet> packet);
  void P2pRxEvent (std::string path, Ptr<const Packet> packet);
  std::list<std::string> Split (std::string input, std::string sep) const;
  uint32_t GetNodeId (std::string path) const;
  uint32_t GetDeviceId (std::string path) const;
  std::pair<uint32_t,uint32_t> P2pGetSrcDst (std::string path) const;

  std::ofstream m_os;
};

NamTracer::NamTracer ()
{}
NamTracer::~NamTracer ()
{
  m_os.close ();
}

std::list<std::string>
NamTracer::Split (std::string input, std::string sep) const
{
  NS_LOG_FUNCTION (this << input << sep);
  std::list<std::string> retval;
  std::string::size_type cur = 0, next;
  while (true)
    {
      next = input.find (sep, cur);
      if (next == cur)
	{
	  cur ++;
	  continue;
	} 
      else if (next == std::string::npos)
	{
	  if (input.size () != cur)
	    {
	      retval.push_back (input.substr (cur, input.size () - cur));
	    }
	  break;
	}
      retval.push_back (input.substr (cur, next - cur));
      cur = next + 1;
    }
  return retval;
}

uint32_t
NamTracer::GetNodeId (std::string path) const
{
  std::list<std::string> items = Split (path, "/");
  NS_ASSERT (items.front () == "NodeList");
  items.pop_front ();
  std::istringstream iss;
  iss.str (items.front ());
  uint32_t node;
  iss >> node;
  return node;
}
uint32_t 
NamTracer::GetDeviceId (std::string path) const
{
  std::list<std::string> items = Split (path, "/");
  NS_ASSERT (items.front () == "NodeList");
  items.pop_front ();
  items.pop_front ();
  NS_ASSERT (items.front () == "DeviceList");
  items.pop_front ();
  std::istringstream iss;
  iss.str (items.front ());
  uint32_t device;
  iss >> device;
  return device;
}
std::pair<uint32_t,uint32_t>
NamTracer::P2pGetSrcDst (std::string path) const
{
  uint32_t nodeid = GetNodeId (path);
  uint32_t deviceid = GetDeviceId (path);
  Ptr<Node> node = NodeList::GetNode (nodeid);
  Ptr<NetDevice> device = node->GetDevice (deviceid);
  Ptr<Channel> channel = device->GetChannel ();
  NS_ASSERT (channel->GetNDevices () == 2);
  Ptr<NetDevice> other;
  if (channel->GetDevice (0) != device)
    {
      other = channel->GetDevice (0);
    }
  else
    {
      other = channel->GetDevice (1);
    }
  return std::make_pair (other->GetNode ()->GetId (), nodeid);
}


void 
NamTracer::P2pEnqueueEvent (std::string path, Ptr<const Packet> packet)
{
  std::pair<uint32_t,uint32_t> srcDst = P2pGetSrcDst (path);
  m_os << "+ -t " << Simulator::Now ().GetSeconds ()
       << " -s " << srcDst.second
       << " -d " << srcDst.first
       << " -e " << packet->GetSize ()
       << " -i " << packet->GetUid ()
       << std::endl;
}
void 
NamTracer::P2pDequeueEvent (std::string path, Ptr<const Packet> packet)
{
  std::pair<uint32_t,uint32_t> srcDst = P2pGetSrcDst (path);
  m_os << "- -t " << Simulator::Now ().GetSeconds ()
       << " -s " << srcDst.second
       << " -d " << srcDst.first
       << " -e " << packet->GetSize ()
       << " -i " << packet->GetUid ()
       << std::endl;
  m_os << "h -t " << Simulator::Now ().GetSeconds ()
       << " -s " << srcDst.second
       << " -d " << srcDst.first
       << " -e " << packet->GetSize ()
       << " -i " << packet->GetUid ()
       << std::endl;
}
void 
NamTracer::P2pDropEvent (std::string path, Ptr<const Packet> packet)
{
  std::pair<uint32_t,uint32_t> srcDst = P2pGetSrcDst (path);
  m_os << "d -t " << Simulator::Now ().GetSeconds ()
       << " -s " << srcDst.second
       << " -d " << srcDst.first
       << " -e " << packet->GetSize ()
       << " -i " << packet->GetUid ()
       << std::endl;
}
void 
NamTracer::P2pRxEvent (std::string path, Ptr<const Packet> packet)
{  
  std::pair<uint32_t,uint32_t> srcDst = P2pGetSrcDst (path);
  m_os << "r -t " << Simulator::Now ().GetSeconds ()
       << " -s " << srcDst.first
       << " -d " << srcDst.second
       << " -e " << packet->GetSize ()
       << " -i " << packet->GetUid ()
       << std::endl;
}


void
NamTracer::SetupP2pTraces (uint32_t nodeid, uint32_t deviceid)
{
  std::ostringstream oss;
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::PointToPointNetDevice/Rx";
  Config::Connect (oss.str (), MakeCallback (&NamTracer::P2pRxEvent, Ptr<NamTracer> (this)));
  oss.str ("");
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::PointToPointNetDevice/TxQueue/Enqueue";
  Config::Connect (oss.str (), MakeCallback (&NamTracer::P2pEnqueueEvent, Ptr<NamTracer> (this)));
  oss.str ("");
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::PointToPointNetDevice/TxQueue/Dequeue";
  Config::Connect (oss.str (), MakeCallback (&NamTracer::P2pDequeueEvent, Ptr<NamTracer> (this)));
  oss.str ("");
  oss << "/NodeList/" << nodeid << "/DeviceList/" << deviceid << "/$ns3::PointToPointNetDevice/TxQueue/Drop";
  Config::Connect (oss.str (), MakeCallback (&NamTracer::P2pDropEvent, Ptr<NamTracer> (this)));
}

void
NamTracer::StartTracing (std::string filename, double width, double height)
{
  m_os.open (filename.c_str ());
  // I tested this code with nam 1.13. Other versions might work.
  m_os << "V -t * -v 1.13 -a 0" << std::endl;
  if (width != 0.0 && height != 0.0)
    {
      m_os << "W -t * -x " << width << " -y " << height << std::endl;
    }
  for (NodeList::Iterator i = NodeList::Begin (); i != NodeList::End (); ++i)
    {
      Ptr<Node> node = *i;
      m_os << "n -t * -a " << node->GetId () << " -s " << node->GetId () << " -S UP -v circle" << std::endl;
    }
  // gather all links.
  std::vector<Ptr<Channel> > links;
  for (NodeList::Iterator i = NodeList::Begin (); i != NodeList::End (); ++i)
    {
      Ptr<Node> node = *i;
      for (uint32_t j = 0; j < node->GetNDevices (); ++j)
	{
	  Ptr<NetDevice> dev = node->GetDevice (j);
	  Ptr<Channel> channel = dev->GetChannel ();
	  if (channel && std::find (links.begin (), links.end (), channel) == links.end ())
	    {
	      links.push_back (channel);
	    }
	}
    }
  uint32_t lanId = NodeList::GetNNodes ();
  for (std::vector<Ptr<Channel> >::const_iterator i = links.begin (); i != links.end (); ++i)
    {
      Ptr<Channel> channel = *i;
      Ptr<PointToPointChannel> p2pChannel = channel->GetObject<PointToPointChannel> ();
      if (p2pChannel != 0)
	{
	  TimeValue delay;
	  p2pChannel->GetAttribute ("Delay", delay);
	  DataRateValue rate;
	  Ptr<PointToPointNetDevice> p2pDev = p2pChannel->GetDevice (0)->GetObject<PointToPointNetDevice> ();
	  p2pDev->GetAttribute ("DataRate", rate);
	  uint32_t a = channel->GetDevice (0)->GetNode ()->GetId ();
	  uint32_t b = channel->GetDevice (1)->GetNode ()->GetId ();
	  m_os << "l -t * -s " << a
	       << " -d " << b
	       << " -S UP -r " << rate.Get ().GetBitRate ()
	       << " -D " << delay.Get ().GetSeconds () 
	       << std::endl;
	  m_os << "q -t * -s " << a << " -d " << b 
	       << " -a 0.5" << std::endl;
	  m_os << "q -t * -s " << b << " -d " << a 
	       << " -a 0.5" << std::endl;
	  SetupP2pTraces (a, channel->GetDevice (0)->GetIfIndex ());
	  SetupP2pTraces (b, channel->GetDevice (1)->GetIfIndex ());
	}
      Ptr<CsmaChannel> csmaChannel = channel->GetObject<CsmaChannel> ();
      if (csmaChannel != 0)
	{
	  TimeValue delay;
	  csmaChannel->GetAttribute ("Delay", delay);
	  DataRateValue rate;
	  csmaChannel->GetAttribute ("DataRate", rate);
	  m_os << "X -t * -n " << lanId 
	     << " -r " << rate.Get ().GetBitRate () 
	     << " -D " << delay.Get ().GetSeconds () 
	     << std::endl;
	  for (uint32_t j = 0; j < csmaChannel->GetNDevices (); ++j)
	    {
	      Ptr<CsmaNetDevice> csmaDev = csmaChannel->GetDevice (j)->GetObject<CsmaNetDevice> ();
	      m_os << "L -t * -s " << lanId << " -d " << csmaDev->GetNode ()->GetId () << std::endl;
	    }
	  lanId++;
	}
    }
}

NamHelper::NamHelper ()
  : m_width (0.0),
    m_height (0.0)
{}

void
NamHelper::StartTracing (std::string filename)
{
  Ptr<NamTracer> tracer = Create<NamTracer> ();
  tracer->StartTracing (filename, m_width, m_height);
}

} // namespace ns3
